import sys
import re
import pprint
from collections import OrderedDict


if len(sys.argv) < 2:
    print ("""
    sat_solver is a program used to read in a set of clauses in satcompetition.org
    and output whether it is satisfiable or not. It is run from the command line
    using the syntax:
    php sat_solver.php filename
    """)

lines = open(sys.argv[1])
clauses = []

for line in lines:
    line = line.strip() # remove beginning and ending white space of the string
    if line[0] == 'c' or line[0] == 'p':
        continue
    pre_clause = re.split('\s+', line)
    clause = []
    for pre_literal in pre_clause:
        literal = int(pre_literal)
        if literal != 0:
            clause.append(literal)
    if len(clause) != 0:
        clauses.append(clause)
print ("Clauses")
print (clauses, "\n") # no pretty print

def dpllSat (clauses):
    symbols = getSymbolsClauses(clauses)
    return dpll(clauses, symbols, {})

diff = lambda l1,l2: [x for x in l1 if x not in l2]

def dpll (clasues, symbols, assignment):
    print ("Symbols")
    print(symbols, "\n") # no pretty print
    print("Assignment")
    print(assignment, "\n")
    if checkAllClausesTrue (clauses, assignment):
        return assignment
    if checkForFalseClause (clasues, symbols, assignment):
        return False
    literal = findPureSymbol(symbols)

    if literal:
        assignment[literal] = 1
        assignment[-literal] = -1
        symbols = diff(symbols, [literal, -literal])
        # print ("What is the list diff? ", symbols)
        return dpll(clasues, symbols, assignment)

    literal = findUnitClause(clauses, assignment)
    if literal:
        assignment[literal] = 1
        assignment[-literal] = -1
        symbols = diff(symbols, [literal, -literal])
        # print ("What is the list diff? ", symbols)

        return dpll(clauses, symbols, assignment)

    literal = choose(symbols)
    symbols = diff(symbols, [literal, -literal])
    # symbols = diff(symbols, [literal, -literal])
    assignment[literal] = 1
    assignment[-literal] = -1
    sat_assignment = dpll(clauses, symbols, assignment)
    if sat_assignment == False:
        assignment[literal] = -1
        assignment[-literal] = 1
        sat_assignment = dpll(clauses, symbols, assignment)
    return sat_assignment

def getSymbolsClauses(clauses):
    symbols = {}
    for clause in clauses:
        for literal in clause:
            symbols[literal] = literal
    # print ('What is symbols', symbols)
    return list(symbols.values())


def checkAllClausesTrue(clauses, assignment):
    for clause in clauses:
        value = False
        for literal in clause:
            # print (assignment[literal])
            try:
                if assignment[literal] and assignment[literal] > 0:
                    value = True
                    break
            except KeyError:
                pass
        if not value:
            return False
    return True

def checkForFalseClause(clauses, symbols, assignment):
    for clause in clauses:
        value = True
        for literal in clause:
            if literal in symbols:
                value = False
                break
            if assignment[literal] > 0:
                value = False
                break
        if value:
            return True
    return False

def findPureSymbol(symbols):
    for literal in symbols:
        if -literal not in symbols:
            return literal
    return False

def findUnitClause(clauses, assignment):
      assigned_literals = assignment.keys()
      for clause in clauses:
          unassigned = diff(clause, assigned_literals)
          if len(unassigned) == 1:
              already_True = False
              for literal in clause:
                  try:
                      if assignment[literal] and assignment[literal] > 0:
                          already_True = True
                          break
                  except KeyError:
                      pass
              if not already_True:
                  return unassigned.pop()   # get and remove last element of a list !!!!
      return False

def choose(symbols):
    return symbols.pop() # get and remove the last element of a list !!!

dpllSa_test = dpllSat(clauses) # var_dump
if(dpllSa_test == False):
    print ("satisfiable?: ", dpllSa_test)
